#include <Array.au3>
#include <File.au3>

$sSciTE_Dir = FileSelectFolder('Select SciTE dir:', '', 0, _PathFull(@AutoItExe & '\..\SciTe'))
If @error Then Exit

$sASH_KwrdsApi_File = @ScriptDir & '\Resources\au3.keywords.properties'
$sAutoIt_KwrdsApi_File = $sSciTE_Dir & '\properties\au3.keywords.properties'
$sAutoIt_Api_File = $sSciTE_Dir & '\api\au3.api'
$aRead_Api_File = FileRead($sAutoIt_Api_File)

_Add_Missing_Keywords('au3.keywords.functions')
_Add_Missing_Keywords('au3.keywords.udfs')
_Add_Missing_Keywords('au3.keywords.keywords')
_Add_Missing_Keywords('au3.keywords.macros')
_Add_Missing_Keywords('au3.keywords.preprocessor')
_Add_Missing_Keywords('au3.keywords.special')
_Add_Missing_Keywords('au3.keywords.sendkeys')

MsgBox(64, 'Add missing keywords', 'Done!')

Func _Add_Missing_Keywords($sKey)
	Local $a_Old_Keywords = _PropertiesRead($sASH_KwrdsApi_File, $sKey)
	
	If @error Then
		ConsoleWrite(@LF & '! Error reading key "' & $sKey & '" in [' & $sASH_KwrdsApi_File & ']' & @LF)
		Return
	EndIf
	
	Local $a_New_Keywords = _PropertiesRead($sAutoIt_KwrdsApi_File, $sKey)
	
	If @error Then
		ConsoleWrite(@LF & '! Error reading key "' & $sKey & '" in [' & $sAutoIt_KwrdsApi_File & ']' & @LF)
		Return
	EndIf
	
	Local $sKeyword, $sKeywords
	Local $aAdd_Kwrds[1] = [0]
	
	For $i = 1 To $a_New_Keywords[0]
		If _ArraySearch($a_Old_Keywords, $a_New_Keywords[$i], 1) > -1 Then
			ContinueLoop
		EndIf
		
		$sKeyword = StringRegExpReplace($aRead_Api_File, '(?si).*?\R\h*(' & $a_New_Keywords[$i] & ')\W.*', '\1')
		
		If @extended And $sKeyword = $a_New_Keywords[$i] Then
			$aAdd_Kwrds[0] += 1
			ReDim $aAdd_Kwrds[$aAdd_Kwrds[0] + 1]
			$aAdd_Kwrds[$aAdd_Kwrds[0]] = $sKeyword
			
			ConsoleWrite(StringFormat('New keyword in (%s): %s', $sKey, $sKeyword) & @LF)
		EndIf
	Next
	
	If $aAdd_Kwrds[0] = 0 Then
		ConsoleWrite(@LF & '> No new keywords for ' & $sKey & @LF)
	Else
		$sKeywords = _ArrayToString($aAdd_Kwrds, '|', 1)
		
		If Not @error Then
			_PropertiesWrite($sASH_KwrdsApi_File, $sKey, $sKeywords)
			ConsoleWrite(@LF & '+ ' & $sKey & ' Keywords added' & @LF)
		Else
			ConsoleWrite(@LF & '! ' & $sKey & ' Keywords NOT added' & @LF)
		EndIf
	EndIf
	
	ConsoleWrite(@LF & '-----------------------------------------------------' & @LF & @LF)
EndFunc

Func _PropertiesWrite($sFile, $sProperty, $sValue, $iPropsPerLine = 5)
	Local $sFileRead = FileRead($sFile)
	Local $aValue = StringSplit(StringReplace($sValue, '\', ''), '|')
	
	$sValue = ''
	
	For $i = 1 To $aValue[0]
		$sValue &= $aValue[$i]
		
		If $i < $aValue[0] Then
			If ($aValue[0] = 1 Or Mod($i, $iPropsPerLine) = 0) Then
				$sValue &= ' \' & @CRLF & @TAB
			Else
				$sValue &= ' '
			EndIf
		EndIf
	Next
	
	$sFileRead = StringRegExpReplace($sFileRead, _
		'(?si)(.*?\R?\Q' & $sProperty & '\E)\s*=\s*(.*)', _
		'\1=' & StringReplace($sValue, '\', '\\') & ' \\' & @CRLF & @TAB & '\2', 1)
	
	$hFile = FileOpen($sFile, 2)
	FileWrite($hFile, $sFileRead)
	Return FileClose($hFile)
EndFunc

Func _PropertiesRead($sFile, $sProperty, $sDefault = '', $iRetArr = 1, $iDelDups = 1)
	Local $aFileRead = StringSplit(StringStripCR(FileRead($sFile)), @LF)
	Local $sRet, $iPropLen = StringLen($sProperty)
	
	For $i = 1 To $aFileRead[0]
		$aFileRead[$i] = StringStripWS($aFileRead[$i], 3)
		
		If StringLeft($aFileRead[$i], $iPropLen + 1) = $sProperty & '=' Then
			$aFileRead[$i] = StringTrimLeft($aFileRead[$i], $iPropLen + 1)
			
			If StringRight($aFileRead[$i], 1) = '\' Then
				$aFileRead[$i] = StringTrimRight($aFileRead[$i], 1)
			EndIf
			
			$sRet &= StringStripWS($aFileRead[$i], 3)
			
			For $j = $i + 1 To $aFileRead[0]
				If StringInStr($aFileRead[$j], '=') Then
					ExitLoop 2
				ElseIf StringLeft($aFileRead[$j], 1) <> '#' Then
					If StringRight($aFileRead[$j], 1) = '\' Then
						$aFileRead[$j] = StringTrimRight($aFileRead[$j], 1)
					EndIf
					
					$sRet &= ' ' & StringStripWS($aFileRead[$j], 3)
				EndIf
			Next
			
			ExitLoop
		EndIf
	Next
	
	If $sRet = '' Then
		$sRet = $sDefault
	EndIf
	
	$sRet = StringStripWS($sRet, 3)
	
	If $iRetArr Then
		$sRet = StringSplit($sRet, ' ' & @TAB)
		
		If $iDelDups Then
			_ArrayDeleteDuplicates($sRet)
		EndIf
	EndIf
	
	Return $sRet
EndFunc

Func _ArrayDeleteDuplicates(ByRef $aArray)
	Local $oDict, $aRet[UBound($aArray)]
	
	$oDict = ObjCreate('Scripting.Dictionary')
	
	If Not IsObj($oDict) Then
		Return SetError(1, 0, 0)
	EndIf
	
	$oDict.CompareMode = 1 ;Text mode (not case sensitive)
	
	For $i = 1 To UBound($aArray)-1
		If Not $oDict.Exists($aArray[$i]) Then
			$oDict.Add($aArray[$i], $aArray[$i])
			
			$aRet[0] += 1
			$aRet[$aRet[0]] = $aArray[$i]
		EndIf
	Next
	
	ReDim $aRet[$aRet[0] + 1]
	$aArray = $aRet
	
	Return 1
EndFunc
